import shutil

import pandas as pd
from stqdm import stqdm
import os
from data_teacher import constants as cnst
import cv2
from data_teacher.differ import dif
from imutils import paths
from torchvision import datasets, transforms
import torch
import numpy as np
import umap
import umap.plot
import plotly.express as px
from sklearn import preprocessing
from PIL import Image
from whylogs import get_or_create_session
import seaborn as sns
import matplotlib.pyplot as plt
from data_teacher import utils

# Sharpening Kernel
sharp_kernel = np.array([[0, -1, 0],
                         [-1, 5, -1],
                         [0, -1, 0]])


class CustomDataset(torch.utils.data.Dataset):
    def __init__(self, dataframe, transform=None):
        self.annotations = dataframe
        self.transform = transform

    def __len__(self):
        return len(self.annotations)

    def __getitem__(self, index):
        img_id = self.annotations.iloc[index, 0]  # Col 1 = filepath
        img = Image.open(img_id).convert("RGB")
        y_label = torch.tensor(float(self.annotations.iloc[index, 3]))  # Col 3 = categorical_label
        # print(f"Image PAth : {img_id}")
        # print(f"Label String: {y_label}")
        if self.transform is not None:
            img = self.transform(img)

        return (img, y_label)


def get_class_count(df):
    grp = df.groupby(["class_label"])['img_path'].nunique()
    count_dict = {key: grp[key] for key in list(grp.keys())}
    return count_dict


def get_class_proportions(df):
    class_counts = get_class_count(df)
    return {val[0]: round(val[1] / df.shape[0], 3) for val in class_counts.items()}


def get_dataset_count_df():
    # dataset_count_df = pd.DataFrame(cnst.dataset_class_data_dict)
    data_df = pd.read_csv(cnst.current_metadata_file)
    dataset_count_df = pd.DataFrame(get_class_count(data_df), index=[0])
    return dataset_count_df


def get_resolution_count_df():
    data_df = pd.read_csv(cnst.current_metadata_file)
    resolution_list = data_df["resolution"].unique()

    resolution_count_dict = {}
    for resolution in resolution_list:
        resolution_count_dict[resolution] = len(data_df.loc[data_df["resolution"] == resolution])
    return pd.DataFrame(resolution_count_dict, index=[0])


def get_aspect_ratio_count_df():
    data_df = pd.read_csv(cnst.current_metadata_file)
    aspect_list = data_df["aspect_ratio"].unique()

    aspect_count_dict = {}
    for ratio in aspect_list:
        aspect_count_dict[f"{ratio}"] = len(data_df.loc[data_df["aspect_ratio"] == ratio])
    return pd.DataFrame(aspect_count_dict, index=[0])


def check_blank_black_img(pil_img):
    extrema = pil_img.convert("L").getextrema()
    if extrema == (0, 0):
        # all black
        return 1
    return 0


def check_blank_white_img(pil_img):
    extrema = pil_img.convert("L").getextrema()
    if extrema == (255, 255):
        # all white
        return 1
    return 0


def get_quality_overview():
    data_df = pd.read_csv(cnst.current_metadata_file)
    quality_data_df = data_df[["plain_black", "plain_white", "Duplicate", "Blurred", "Good"]]
    # quality_data_df.columns = ["Count"]
    quality_dict = {}
    for column in quality_data_df.columns:
        # if column is "plain_black" or "plain_white":
        quality_dict[column] = len(quality_data_df.loc[quality_data_df[column] == 1])
        # else:
        #     quality_dict[column] = 0
    return pd.DataFrame(quality_dict, index=[0])


def check_blur_img(img_path, threshold=100):
    image = cv2.imread(img_path)
    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    fm = cv2.Laplacian(gray, cv2.CV_64F).var()
    if fm < threshold:
        # blurry
        return 1
    return 0


def check_duplicate_img(dataset_path, resize_px=50, streamlit_show=False):
    # to check duplicate images within class directory structure
    duplicate_image_list = dif.compare_images(dataset_path, similarity='high', px_size=50, streamlit=streamlit_show)
    return duplicate_image_list


def sharpen_image(img_path):
    print(f"Sharpening image: {img_path}")
    # load the image into system memory
    image = cv2.imread(img_path, flags=cv2.COLOR_BGR2GRAY)
    image_sharp = cv2.filter2D(src=image, ddepth=-1, kernel=sharp_kernel)
    cv2.imwrite(img_path, image_sharp)


def get_infer_df(csv_path):
    pred_df = pd.read_csv(csv_path)
    columns = pred_df.columns

    infer_summary_dict = {}
    ground_labels = pred_df['Ground Truth'].unique()

    # Show the True Positive Count for each class
    for column_name in columns:
        if column_name != "image_path" and column_name != "Ground Truth":
            column_key = column_name + ' (True Positive)'
            infer_summary_dict[column_key] = {}
            for gt in ground_labels:
                infer_summary_dict[column_key][gt] = 0
            df = pred_df[["Ground Truth", column_name]]
            # display(df.head(5))
            for index, row in df.iterrows():
                if row["Ground Truth"] == row[column_name]:
                    infer_summary_dict[column_key][row["Ground Truth"]] += 1
        if column_name == "Ground Truth":
            infer_summary_dict[column_name] = {}
            for gt in ground_labels:
                infer_summary_dict[column_name][gt] = len(pred_df.loc[pred_df[column_name] == gt])
    return pd.DataFrame(infer_summary_dict).T


def extract_classwise_feature_embeddings(dataset_path, streamlit_show=False):
    image_list = sorted(list(paths.list_images(dataset_path)))
    # create dataloader with required transforms
    tc = transforms.Compose([
        transforms.Resize((100, 100)),
        transforms.ToTensor()
    ])
    image_datasets = datasets.ImageFolder(dataset_path, transform=tc)
    num_classes = len(image_datasets.classes)

    dloader = torch.utils.data.DataLoader(image_datasets, shuffle=False)

    # fetch pretrained model
    model = torch.hub.load('pytorch/vision:v0.10.0', 'resnet50', pretrained=True)
    # Select the desired layer
    layer = model._modules.get('avgpool')

    def copy_embeddings(m, i, o):
        """Copy embeddings from the penultimate layer.
        """
        o = o[:, :, 0, 0].detach().numpy().tolist()
        outputs.append(o)

    outputs = []

    # attach hook to the penultimate layer
    _ = layer.register_forward_hook(copy_embeddings)

    model.eval()  # Inference mode

    # Generate image's embeddings for all images in dloader and saves
    # them in the list outputs
    for X, y in stqdm(dloader):
        _ = model(X)

    print(f"Number of embeddings: {len(outputs)}")

    # flatten list of embeddings to remove batches
    list_embeddings = [item for sublist in outputs for item in sublist]

    label_list = []
    for img, label in dloader:
        label_list.append(label.item())

    # Create embedding from the dataset
    umap_2d = umap.UMAP(n_components=3, random_state=42)
    #                     n_neighbors=5,
    #                     min_dist=0.3,
    #                     metric='euclidean')

    umap_2d.fit(list_embeddings)

    projections = umap_2d.transform(list_embeddings)

    print(np.array(projections).shape)

    fig = px.scatter_3d(
        projections, title="Feature wise Class Distribution",
        x=0, y=1, z=2,
        color=list(map(lambda x: image_datasets.classes[x], label_list)),
        labels={'color': 'classes', '0': 'image_embedding_val_0', '1': 'image_embedding_val_1'},
        symbol=list(map(lambda x: image_datasets.classes[x], label_list))
    )

    return fig


def extract_splitwise_feature_embeddings(metadata_csv, streamlit_show=False):
    # create dataloader with required transforms
    tc = transforms.Compose([
        transforms.Resize((100, 100)),
        transforms.ToTensor()
    ])

    main_dataset_df = pd.read_csv(metadata_csv)

    print(f"There are {len(main_dataset_df)} images in the dataset.")
    main_dataset_df = main_dataset_df[["img_path", "class_label", "split_type"]]

    class_list = main_dataset_df["class_label"].unique()
    split_type_list = main_dataset_df["split_type"].unique()
    split_type_list.sort()

    # fetch pretrained model
    model = torch.hub.load('pytorch/vision:v0.10.0', 'resnet50', pretrained=True)
    # Select the desired layer
    layer = model._modules.get('avgpool')

    def copy_embeddings(m, i, o):
        """Copy embeddings from the penultimate layer.
        """
        o = o[:, :, 0, 0].detach().numpy().tolist()
        outputs.append(o)

    outputs = []

    # attach hook to the penulimate layer
    _ = layer.register_forward_hook(copy_embeddings)

    model.eval()  # Inference mode

    class_wise_split_dict = {}
    for class_name in class_list:
        class_wise_split_dict[class_name] = main_dataset_df.groupby("class_label").get_group(class_name)

    split_wise_scatter_plot_dict = {}
    for class_name in stqdm(class_list):
        print(f"Processing for the class : {class_name}")
        outputs = []

        le = preprocessing.LabelEncoder()
        le.fit(class_wise_split_dict[class_name].split_type)
        class_wise_split_dict[class_name]['categorical_label'] = le.transform(
            class_wise_split_dict[class_name].split_type)

        class_dataset = CustomDataset(class_wise_split_dict[class_name], transform=tc)

        class_dloader = torch.utils.data.DataLoader(class_dataset, shuffle=False)
        print(f"Length of the DataLoader for {class_name} is : {len(class_dloader)}")

        # Generate image's embeddings for all images in dloader and saves
        # them in the list outputs
        for X, y in class_dloader:
            _ = model(X)

        print(len(outputs))

        # flatten list of embeddings to remove batches
        list_embeddings = [item for sublist in outputs for item in sublist]
        label_list = []
        for img, label in class_dloader:
            label_list.append(label.item())

        # Create embedding from the dataset
        umap_2d = umap.UMAP(random_state=42)
        # n_neighbors=5,
        # min_dist=0.3,
        # metric='euclidean')

        umap_2d.fit(list_embeddings)

        projections = umap_2d.transform(list_embeddings)

        print(np.array(projections).shape)

        fig = px.scatter(
            projections, title=f"Split wise Data Distribution for {class_name}",
            x=1, y=0,
            color=list(map(lambda x: split_type_list[int(x)], label_list)),
            labels={'color': 'split type', '0': 'image_embedding_val_0', '1': 'image_embedding_val_1'},
            symbol=list(map(lambda x: split_type_list[int(x)], label_list))
        )
        split_wise_scatter_plot_dict[class_name] = fig
        # fig.show()
    return split_wise_scatter_plot_dict


# Optimize the dataset based on recommendations - remove black/white/duplicate images, etc
def optimize_dataset(action_dict, dataset_dir, metadata_csv):
    print(f"Task to perform: {action_dict}")
    data_df = pd.read_csv(metadata_csv)

    orig_dir_path = f"{dataset_dir}_orig_{utils.get_date_time_str()}"

    # Copy whole data from dataset directory to new one
    print(f"Source Path: {dataset_dir}")
    print(f"Destination Path: {orig_dir_path}")
    utils.copy_whole_dir_tree(dataset_dir, orig_dir_path)

    # Remove erroneous images from old directory
    if "remove_erroneous" in action_dict:
        if action_dict["remove_erroneous"]:
            black_df = data_df.loc[data_df["plain_black"] == 1]

            for index, row in black_df.iterrows():
                data_df.drop([index], inplace=True)
                if utils.is_exist(row["img_path"]):
                    print(f"Removing : {row['img_path']}")
                    os.remove(row["img_path"])

            print(f"New Dataframe Length after removing black images: {len(data_df)}")

            white_df = data_df.loc[data_df["plain_white"] == 1]
            for index, row in white_df.iterrows():
                data_df.drop([index], inplace=True)
                if utils.is_exist(row["img_path"]):
                    print(f"Removing : {row['img_path']}")
                    os.remove(row["img_path"])

            print(f"New Dataframe Length after removing white images: {len(data_df)}")

    # Remove duplicate images from old directory
    if "remove_duplicates" in action_dict:
        if action_dict["remove_duplicates"]:
            duplicate_df = data_df.loc[data_df["Duplicate"] == 1]

            for index, row in duplicate_df.iterrows():
                data_df.drop([index], inplace=True)
                if utils.is_exist(row["img_path"]):
                    print(f"Removing : {row['img_path']}")
                    os.remove(row["img_path"])

            print(f"New Dataframe Length after removing duplicate images: {len(data_df)}")

    # Re-store images to old directory after deblurring
    if "deblur_images" in action_dict:
        if action_dict["deblur_images"]:
            blurred_df = data_df.loc[data_df["Blurred"] == 1]
            for index, row in blurred_df.iterrows():
                if utils.is_exist(row["img_path"]):
                    data_df.at[index, 'Blurred'] = 0
                    sharpen_image(row['img_path'])
            print(f"New Dataframe Length after sharpening blurred images: {len(data_df)}")

    data_df.to_csv(metadata_csv)


# TODO - change this function to plot drift scatter graph for train data from dataset and infer data path
def data_drift_feature_embeddings(streamlit_show=False, infer_data_path=""):
    # create dataloader with required transforms
    tc = transforms.Compose([
        transforms.Resize((100, 100)),
        transforms.ToTensor()
    ])

    main_dataset_df = pd.read_csv(cnst.current_metadata_file)

    print(f"There are {len(main_dataset_df)} images in the dataset.")
    main_dataset_df = main_dataset_df[["img_path", "class_label", "split_type"]]

    class_list = main_dataset_df["class_label"].unique()
    split_type_list = main_dataset_df["split_type"].unique()
    split_type_list.sort()

    # fetch pretrained model
    model = torch.hub.load('pytorch/vision:v0.10.0', 'resnet50', pretrained=True)
    # Select the desired layer
    layer = model._modules.get('avgpool')

    def copy_embeddings(m, i, o):
        """Copy embeddings from the penultimate layer.
        """
        o = o[:, :, 0, 0].detach().numpy().tolist()
        outputs.append(o)

    outputs = []

    # attach hook to the penultimate layer
    _ = layer.register_forward_hook(copy_embeddings)

    model.eval()  # Inference mode

    class_wise_split_dict = {}
    for class_name in class_list:
        class_wise_split_dict[class_name] = main_dataset_df.groupby("class_label").get_group(class_name)

    split_wise_scatter_plot_dict = {}
    for class_name in stqdm(class_list):
        print(f"Processing for the class : {class_name}")
        outputs = []

        le = preprocessing.LabelEncoder()
        le.fit(class_wise_split_dict[class_name].split_type)
        class_wise_split_dict[class_name]['categorical_label'] = le.transform(
            class_wise_split_dict[class_name].split_type)

        class_dataset = CustomDataset(class_wise_split_dict[class_name], transform=tc)

        class_dloader = torch.utils.data.DataLoader(class_dataset, shuffle=False)
        print(f"Length of the DataLoader for {class_name} is : {len(class_dloader)}")

        # Generate image's embeddings for all images in dloader and saves
        # them in the list outputs
        for X, y in class_dloader:
            _ = model(X)

        print(len(outputs))

        # flatten list of embeddings to remove batches
        list_embeddings = [item for sublist in outputs for item in sublist]
        label_list = []
        for img, label in class_dloader:
            label_list.append(label.item())

        # Create embedding from the dataset
        umap_2d = umap.UMAP(random_state=0)
        # n_neighbors=5,
        # min_dist=0.3,
        # metric='euclidean')

        umap_2d.fit(list_embeddings)

        projections = umap_2d.transform(list_embeddings)

        print(np.array(projections).shape)

        fig = px.scatter(
            projections, title=f"Split wise Data Distribution for {class_name}",
            x=1, y=0,
            color=list(map(lambda x: split_type_list[int(x)], label_list)),
            labels={'color': 'split type'},
            symbol=list(map(lambda x: split_type_list[int(x)], label_list))
        )
        split_wise_scatter_plot_dict[class_name] = fig
        # fig.show()
    return split_wise_scatter_plot_dict


def create_logger(dir_path, partition_type="train", session=None):
    if partition_type == "train":
        dir_path = os.path.join(dir_path, partition_type)

    logger = session.logger(f"{partition_type}_dataset")
    if partition_type == "train":
        for class_name in os.listdir(dir_path):
            for image_path in paths.list_images(os.path.join(dir_path, class_name)):
                logger.log_image(image_path)
    else:
        for image_name in os.listdir(dir_path):
            image_path = os.path.join(dir_path, image_name)
            # print(image_path)
            logger.log_image(image_path)
    return logger


def get_custom_histogram_info(variable, n_bins, train_summary, infer_summary, train_profile, infer_profile):
    min_range = min(train_summary[train_summary["column"] == variable]["min"].values[0],
                    infer_summary[infer_summary["column"] == variable]["min"].values[0])
    max_range = max(train_summary[train_summary["column"] == variable]["max"].values[0],
                    infer_summary[infer_summary["column"] == variable]["max"].values[0])
    min_bins = int((max_range - min_range) / n_bins)
    # print(f"Min : {min_range}")
    # print(f"Min : {max_range}")
    if min_range == 0.0 and max_range == 0.0:
        return None, None, None
    bins = range(int(min_range), int(max_range), min_bins if min_bins != 0 else 1)

    training_counts = np.array(
        train_profile.columns[variable].number_tracker.histogram.get_pmf(bins[:-1]))
    infer_counts = np.array(
        infer_profile.columns[variable].number_tracker.histogram.get_pmf(bins[:-1]))

    return bins, training_counts, infer_counts


def plot_distribution_shift(variable, n_bins, train_summary, infer_summary, train_profile, infer_profile):
    """Visualization for distribution shift"""
    bins, training_counts, infer_counts = get_custom_histogram_info(variable, n_bins, train_summary, infer_summary,
                                                                    train_profile, infer_profile)

    if bins == None:
        return None
    fig, ax = plt.subplots(figsize=(10, 3))
    sns.histplot(x=bins, weights=training_counts, bins=n_bins,
                 label="Training data", color="teal", alpha=0.7, ax=ax, stat='probability')
    sns.histplot(x=bins, weights=infer_counts, bins=n_bins,
                 label="Real-Time data", color="gold", alpha=0.7, ax=ax, stat='probability')
    ax.legend()
    plt.title(f"Distribution of {variable.split('.')[0]}")
    # plt.show()
    return fig


def monitor_data_drift(splitted_data_dir, infer_data_dir):
    _session = None
    session = get_or_create_session()

    train_logger = create_logger(splitted_data_dir, "train", session)
    infer_logger = create_logger(infer_data_dir, "infer", session)

    train_profile = train_logger.profile
    infer_profile = infer_logger.profile

    train_histograms = train_profile.flat_summary()["hist"]
    infer_histograms = infer_profile.flat_summary()["hist"]

    train_summary = train_profile.flat_summary()["summary"]
    infer_summary = infer_profile.flat_summary()["summary"]

    brightness_drift_plot = plot_distribution_shift("Brightness.mean", 30, train_summary, infer_summary, train_profile,
                                                    infer_profile)
    saturation_drift_plot = plot_distribution_shift("Saturation.mean", 30, train_summary, infer_summary, train_profile,
                                                    infer_profile)
    hue_drift_plot = plot_distribution_shift("Hue.mean", 30, train_summary, infer_summary, train_profile,
                                             infer_profile)

    session.close()
    return brightness_drift_plot, saturation_drift_plot, hue_drift_plot
