"""
Copyright (c) 2021 Elise Landman

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
"""

import skimage.color
# import matplotlib.pyplot as plt
import numpy as np
import cv2
import os
import imghdr
import time
from imutils import paths
from tqdm import tqdm
from stqdm import stqdm
""" 
Duplicate Image Finder (DIF): function that searches a given directory for images and finds duplicate/similar images 
among them.
Outputs the number of found duplicate/similar image pairs with a list of the filenames having lower resolution.
"""


class dif:
    def compare_images(dir_path, show_imgs=False, similarity="normal", px_size=50, delete=False, streamlit=False):
        """
        dir_path (str)......folder path to search for duplicate/similar images
        show_imgs (bool).....False = omits the output and doesn't show found images
                             True = shows duplicate/similar images found in output
        similarity (str)....."normal" = searches for duplicates, recommended setting, MSE < 200
                             "high" = searches for exact duplicates, extremly sensitive to details, MSE < 0.1
                             "low" = searches for similar images, MSE < 1000
        px_size (int)........recommended not to change default value
                             resize images to px_size height x width (in pixels) before being compared
                             the higher the pixel size, the more computational ressources and time required
        delete (bool)........! please use with care, as this cannot be undone
                             lower resolution duplicate images that were found are automatically deleted
        streamlit (bool).....Show progress bar in streamlit webpage if True else normal progress bar on terminal output

        OUTPUT (set).........a set of the filenames of the lower resolution duplicate images
        """
        # list where the found duplicate/similar images are stored
        duplicates = []
        lower_res = []

        imgs_matrix = dif.create_imgs_matrix(dir_path, px_size)

        # search for similar images, MSE < 1000
        if similarity == "low":
            ref = 1000
        # search for exact duplicate images, extremly sensitive, MSE < 0.1
        elif similarity == "high":
            ref = 0.1
        # normal, search for duplicates, recommended, MSE < 200
        else:
            ref = 200

        main_img = 0
        compared_img = 1
        nrows, ncols = px_size, px_size
        srow_A = 0
        erow_A = nrows
        srow_B = erow_A
        erow_B = srow_B + nrows

        if streamlit:
            pbar = stqdm(total=imgs_matrix.shape[0] / ncols)
        else:
            pbar = tqdm(total=imgs_matrix.shape[0] / ncols)
        while erow_B <= imgs_matrix.shape[0]:
            pbar.update(1)
            while compared_img < (len(image_files)):
                # select two images from imgs_matrix
                imgA = imgs_matrix[srow_A: erow_A,  # rows
                       0: ncols]  # columns
                imgB = imgs_matrix[srow_B: erow_B,  # rows
                       0: ncols]  # columns
                # compare the images
                rotations = 0
                while image_files[main_img] not in duplicates and rotations <= 3:

                    if rotations != 0:
                        imgB = dif.rotate_img(imgB)
                    err = dif.mse(imgA, imgB)
                    if err < ref:
                        if show_imgs:
                            # dif.show_img_figs(imgA, imgB, err)
                            dif.show_file_info(compared_img, main_img)
                        dif.add_to_list(image_files[main_img], duplicates)
                        dif.check_img_quality(dir_path, image_files[main_img], image_files[compared_img], lower_res)
                    rotations += 1
                srow_B += nrows
                erow_B += nrows
                compared_img += 1

            srow_A += nrows
            erow_A += nrows
            srow_B = erow_A
            erow_B = srow_B + nrows
            main_img += 1
            compared_img = main_img + 1
        pbar.close()
        msg = "\n***\nFound " + str(len(duplicates)) + " duplicate image pairs in " + str(
            len(image_files)) + " total images.\n\nThe following files have lower resolution:"
        # print(msg)
        # print(lower_res, "\n")
        time.sleep(0.5)

        # if delete:
        #     usr = input("Are you sure you want to delete all lower resolution duplicate images? (y/n)")
        #     if str(usr) == "y":
        #         dif.delete_imgs(dir_path, set(lower_res))
        #     else:
        #         print("Image deletion canceled.")
        #         return set(lower_res)
        # else:
        return lower_res

    def _process_directory(dir_path):
        dir_path += os.sep
        if not os.path.isdir(dir_path):
            raise FileNotFoundError(f"Directory: " + dir_path + " does not exist")
        return dir_path

    # Function that searches the folder for image files, converts them to a matrix
    def create_imgs_matrix(dir_path, px_size):
        dir_path = dif._process_directory(dir_path)
        global image_files
        image_files = []
        # create list of all files in directory
        # folder_files = [filename for filename in os.listdir(dir_path)]

        # create images matrix
        counter = 0
        # for filename in folder_files:
        #     if not os.path.isdir(dir_path + filename) and imghdr.what(dir_path + filename):
        class_dir_list = os.listdir(dir_path)
        for class_dir in class_dir_list:
            filename_list = sorted(list(paths.list_images(os.path.join(dir_path, class_dir))))
            for file_path in filename_list:
                # =================================================================================
                if imghdr.what(file_path):
                    # Image file path
                    img = cv2.imdecode(np.fromfile(file_path, dtype=np.uint8), cv2.IMREAD_UNCHANGED)
                    if type(img) == np.ndarray:
                        img = img[..., 0:3]
                        img = cv2.resize(img, dsize=(px_size, px_size), interpolation=cv2.INTER_CUBIC)
                        if len(img.shape) == 2:
                            img = skimage.color.gray2rgb(img)
                        if counter == 0:
                            imgs_matrix = img
                            image_files.append(file_path)
                            counter += 1
                        else:
                            imgs_matrix = np.concatenate((imgs_matrix, img))
                            image_files.append(file_path)
        return imgs_matrix

    # Function that calculates the mean squared error (mse) between two image matrices
    def mse(imageA, imageB):
        err = np.sum((imageA.astype("float") - imageB.astype("float")) ** 2)
        err /= float(imageA.shape[0] * imageA.shape[1])
        return err

    # Function that plots two compared image files and their mse
    # def show_img_figs(imageA, imageB, err):
    #     fig = plt.figure()
    #     plt.suptitle("MSE: %.2f" % (err))
    #     # plot first image
    #     ax = fig.add_subplot(1, 2, 1)
    #     plt.imshow(imageA, cmap=plt.cm.gray)
    #     plt.axis("off")
    #     # plot second image
    #     ax = fig.add_subplot(1, 2, 2)
    #     plt.imshow(imageB, cmap=plt.cm.gray)
    #     plt.axis("off")
    #     # show the images
    #     plt.show()

    # Function for rotating an image matrix by a 90 degree angle
    def rotate_img(image):
        image = np.rot90(image, k=1, axes=(0, 1))
        return image

    # Function for printing filename info of plotted image files
    def show_file_info(compared_img, main_img):
        print("Duplicate file: " + image_files[main_img] + " and " + image_files[compared_img])

    # Function for appending items to a list
    def add_to_list(filename, list):
        list.append(filename)

    # Function for checking the quality of compared images, appends the lower quality image to the list
    def check_img_quality(dir_path, imageA, imageB, list):
        dir_path = dif._process_directory(dir_path)
        size_imgA = os.stat(imageA).st_size
        size_imgB = os.stat(imageB).st_size

        # storing the file_paths in dictionary
        if size_imgA > size_imgB:
            data_dict = {"large": imageA, "small": imageB}
        else:
            data_dict = {"large": imageB, "small": imageA}
        dif.add_to_list(data_dict, list)

    # def delete_imgs(dir_path, filenames_set):
    #     dir_path = dif._process_directory(dir_path)
    #     print("\nDeletion in progress...")
    #     deleted = 0
    #     for filename in filenames_set:
    #         try:
    #             os.remove(filename)
    #             print("Deleted file:", filename)
    #             deleted += 1
    #         except:
    #             print("Could not delete file:", filename)
    #     print("\n***\nDeleted", deleted, "duplicates.")
