import os.path
import pickle
import shutil
from pathlib import Path

from imutils import paths

from data_teacher import analyzer
from helper import constants as cnst, utils
import numpy as np
from PIL import Image
import pandas as pd
from sklearn.model_selection import train_test_split
import time
import streamlit as st
from stqdm import stqdm
from tqdm import tqdm
# from data_teacher.feature_extractor import FeatureExtractor


def load_dataset(streamlit_show=False, dataset_path=None):
    class_dir_list = os.listdir(dataset_path)

    # create a dataframe to save metadata into CSV file
    # metadata single row data format
    # ["img_path", "class_label", "resolution", "aspect_ratio", "plain_black", "plain_white", "Duplicate", "Blurred"]
    metadata_arr = []

    # Find duplicate images from whole dataset
    with st.spinner("Finding Duplicates in Dataset..."):
        start_time = time.process_time()
        duplicate_file_list = analyzer.check_duplicate_img(dataset_path, resize_px=25,
                                                           streamlit_show=streamlit_show)
        end_time = time.process_time() - start_time
        print(f"Time taken to find duplicates: {end_time} seconds")
        small_duplicate_file_list = [data_dict["small"] for data_dict in duplicate_file_list]

    # with st.spinner("Creating feature embeddings..."):
    #     start_time = time.process_time()
    #
    #
    #     end_time = time.process_time() - start_time
    #     print(f"Time taken to find duplicates: {end_time} seconds")
    filename_list = sorted(list(paths.list_images(dataset_path)))
    with st.spinner("Creating classwise metadata of dataset in CSV file..."):
        if streamlit_show:
            pbar = stqdm(total=len(filename_list))
        else:
            pbar = tqdm(total=len(filename_list))
        for file_path in filename_list:
            pbar.update(1)
            class_name = file_path.split(os.path.sep)[-2]
            pil_img = Image.open(file_path)
            im_w, im_h = pil_img.size
            aspect_ratio = im_w / im_h
            is_blurred = 0
            is_duplicate = 0
            is_good = 0
            is_plain_black = analyzer.check_blank_black_img(pil_img)
            is_plain_white = analyzer.check_blank_white_img(pil_img)

            if is_plain_white == 0 and is_plain_black == 0:
                is_blurred = analyzer.check_blur_img(file_path)

            if file_path in small_duplicate_file_list:
                is_duplicate = 1

            if is_plain_white == 0 and is_plain_black == 0 and is_duplicate == 0 and is_blurred == 0:
                is_good = 1
            tmp_arr = [file_path, class_name, f"{im_w}x{im_h}", aspect_ratio, is_plain_black, is_plain_white,
                       is_duplicate, is_blurred, is_good]
            metadata_arr.append(tmp_arr)
        pbar.close()
    data_df = pd.DataFrame(np.array(metadata_arr), columns=["img_path", "class_label", "resolution", "aspect_ratio",
                                                            "plain_black", "plain_white", "Duplicate", "Blurred", "Good"])

    print(f"Total Length of Data : {len(data_df)}")
    print(f"Class Proportions : {analyzer.get_class_proportions(data_df)}")
    time_str = utils.get_date_time_str()
    cnst.current_metadata_file = f"./dataset_metadata_{time_str}.csv"

    # with open(f'embedding_data_{time_str}.pickle', 'wb') as handle:
    #     pickle.dump(embedding_dict, handle, protocol=pickle.HIGHEST_PROTOCOL)

    data_df.to_csv(cnst.current_metadata_file, index=False)
    return cnst.current_metadata_file


def split_dataset(metadata_file, streamlit_show=False):
    try:
        print(f"Current Metadata File : {metadata_file}")
        if not os.path.exists(metadata_file):
            return False, "", ""
        data_df = pd.read_csv(metadata_file)
        train, test_tmp = train_test_split(data_df, test_size=0.3, stratify=data_df['class_label'], random_state=4)

        train_class_proportions = analyzer.get_class_proportions(train)
        test_tmp_class_proportions = analyzer.get_class_proportions(test_tmp)

        test, valid = train_test_split(test_tmp, test_size=0.5, stratify=test_tmp['class_label'], random_state=4)

        test_class_proportions = analyzer.get_class_proportions(test)
        valid_class_proportions = analyzer.get_class_proportions(valid)

        print(f"Train Data Class Proportions : {train_class_proportions}")
        print(f"Test Data Class Proportions : {test_class_proportions}")
        print(f"Valid Data Class Proportions : {valid_class_proportions}")

        split_type_dict = {"train": train,
                           "test": test,
                           "valid": valid}

        merged_metadata_df = pd.DataFrame()

        for split_type in split_type_dict.keys():
            split_type_dict[split_type]["split_type"] = [split_type for i in range(len(split_type_dict[split_type]))]
            merged_metadata_df = pd.concat([merged_metadata_df, split_type_dict[split_type]])

        datetime_string = utils.get_date_time_str()

        # Remove old metadata CSV file
        # os.remove(metadata_file)
        metadata_file = f"./split_dataset_metadata_{datetime_string}.csv"

        # TODO - change root path to current working directory
        target_root_dir = rf"./dataset_split_{datetime_string}"
        # target_root_dir = rf"C:/Users/agajerax/Downloads/dataset_split_{datetime_string}"

        target_root_dir = os.path.abspath(target_root_dir)

        with st.spinner("Splitting the dataset for train, test and validation..."):
            for index, row in merged_metadata_df.iterrows():
                img_path = row["img_path"]
                class_name = row["class_label"]
                split_dir = row["split_type"]
                dir_path = os.path.join(target_root_dir, split_dir, class_name)
                filename = os.path.basename(img_path)
                merged_metadata_df.at[index, 'img_path'] = os.path.join(dir_path, filename)

                # TODO - Uncomment 2 lines for create new different directory of dataset after split
                Path(dir_path).mkdir(parents=True, exist_ok=True)
                shutil.copy2(img_path, dir_path)

            merged_metadata_df.to_csv(metadata_file, index=False)
    except Exception as e:
        return False, "", "", ""
    return True, target_root_dir, metadata_file
